#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <io.h>
#include <ctype.h>
#include <assert.h>
#include <math.h>
#include <string>
#include <vector>
#include <iostream>
#include <algorithm>
#include <scriptol/libsol.hpp>

//	Simple Map - XML, text or HTML Site Map  Generator
//	A free open-source Scriptol program,  by (c) Denis G. Sureau
//	http://www.scriptol.com/seo/      ...the webmaster's source.
#include "path.hpp"
#include "libphp.hpp"
#include "tools.hpp"
#include "options.hpp"
#include "url.hpp"
#include "page.hpp"
#include "parser.hpp"

array siteList=array();
dict siteTime=array();
array parsedList=array();
array skipped=array();
text pageText="";
array siteMap=array();

void output(text t)
{
    siteMap.push(dyn(t));
    return;
}

// get the part of tag inside quotes or double-quotes, thus the url

text extractLink(int off)
{
    text c="";
    text link="";
    while(off<pageText.length())
    {
        c=(pageText[off]);
        if(c=="=")
        {
            goto _lab_parser6;
        }
        if(c==" ")
        {
            goto _lab_parser6;
        }
        if(c=="\"")
        {
            off+=1;
            goto _lab_parser7;
        }
        if(c=="\\'")
        {
            off+=1;
            goto _lab_parser7;
        }
        goto _lab_parser7;
    _lab_parser6:;
        off+=1;
    }
    _lab_parser7:;
    while(off<pageText.length())
    {
        c=(pageText[off]);
        if(c=="\\'")
        {
            goto _lab_parser9;
        }
        if(c=="\"")
        {
            goto _lab_parser9;
        }
        if(c==">")
        {
            goto _lab_parser9;
        }
        link+=c;
        off+=1;
    }
    _lab_parser9:;
    return link.trim();
}

// Test if page already scanned

bool inList(text page)
{
    if(!hasProtocol(page))
    {
        page=Path::merge(siteURL,page);
    }
    if(in_array(dyn(page),parsedList))
    {
        return true;
    }
    return false;
}

// Add page to list, even if not valid, to not scan it again

void addList(text page)
{
    if(!hasProtocol(page))
    {
        page=Path::merge(siteURL,page);
    }
    if(in_array(dyn(page),parsedList))
    {
        return;
    }
    parsedList.push(dyn(page));
    return;
}

// add local file with path to list of all scanned pages

bool addPage(text fname)
{
    fname=fname.trim();
    if(in_array(dyn(fname),parsedList))
    {
        if(DEBUG&&VERBOSE)
        {
            std::cout << fname<< " " << "already parsed"<< "\n";
        }
        return false;
    }
    parsedList.push(dyn(fname));
    if(DEBUG==true)
    {
        std::cout << "Scanning page"<< " " << fname<< "\n";
    }
    return true;
}

// and to the final list of all pages in the site to be referenced
// fname is the URL

bool addLink(text fname)
{
    fname=fname.trim();
    if(!validExtension(fname))
    {
        return false;
    }
    if(!hasProtocol(fname))
    {
        fname=Path::merge(siteURL,fname);
    }
    text url=setURL(fname);
    if(in_array(dyn(url),siteList))
    {
        return false;
    }
    PageWeb *pw=new PageWeb();
    text localname="";
    bool flag=0;
    array _I_1=localPath(fname);
    localname=_I_1.begin().toText();
    flag=_I_1.inc().toBoolean();
    if(in_array(dyn(localname),skipped))
    {
        return false;
    }
    pw->load(localname);
    if(!pw->indexable())
    {
        std::cout << localname<< " " << "skipped, NOINDEX"<< "\n";
        skipped.push(dyn(localname));
        return false;
    }
    siteList.push(dyn(url));
    std::cout << "Added link"<< " " << url<< "\n";
    return true;
}


void addTime(text fname,double lmod)
{
    fname=fname.trim();
    if(!hasProtocol(fname))
    {
        fname=Path::merge(siteURL,fname);
    }
    if((siteTime.at(fname))!=false)
    {
        return;
    }
    if(!validExtension(fname))
    {
        return;
    }
    siteTime[fname]=date("",intval(lmod));
    if(DEBUG==true)
    {
        std::cout << "Added time"<< " " << fname<< " " << date("",intval(lmod))<< "\n";
    }
    return;
}

// Get the URL from the text, skipping other chars

array getLinks(const text &localBase,const text &fname)
{
    array x=array();
    text localLink="";
    x.load(text(basePath+fname).data);
    array pageLinks=array();
    x.reset();
    do
    {
        x[x.key()]=x.value().toText().lower().rTrim();
    }
    while(x.bound());
    pageText=x.join("");
    int offset=0;
    int srcoff=0;
    int shifting=0;
    while(true)
    {
        if(offset!=-1)
        {
            offset=pageText.find("href=",shifting);
        }
        if(srcoff!=-1)
        {
            srcoff=pageText.find("frame src",shifting);
            if(srcoff==-1)
            {
                srcoff=pageText.find("frame  src",shifting);
                if(srcoff!=-1)
                {
                    srcoff+=6;
                }
            }
            else
            {
                srcoff+=6;
            }
        }
        if(offset==-1)
        {
            if(srcoff==-1)
            {
                goto _lab_parser11;
            }
            shifting=srcoff;
        }
        else
        {
            if(srcoff==-1)
            {
                shifting=offset;
            }
            else
            {
                if(offset<srcoff)
                {
                    shifting=offset;
                }
                else
                {
                    shifting=srcoff;
                }
            }
        }
        shifting+=4;
        text link=extractLink(shifting);
        shifting+=link.length();
        if(link=="")
        {
            goto _lab_parser10;
        }
        if((link[0])=="#")
        {
            goto _lab_parser10;
        }
        if(in_array(dyn(link),siteList))
        {
            goto _lab_parser10;
        }
        if(in_array(dyn(link),pageLinks))
        {
            goto _lab_parser10;
        }
        if(in_sequence("../",link.c_str()))
        {
            display(text(" ../ such path is not valid, use absolute URL instead (in ")+link+").");
            goto _lab_parser10;
        }
        if(hasProtocol(link))
        {
            text host=getURL(link);
            if(host==siteURL)
            {
                pageLinks.push(dyn(link));
                addLink(link);
            }
            goto _lab_parser10;
        }
        if(endWith(getcwd(),localBase))
        {
            localLink=link;
        }
        else
        {
            localLink=Path::merge(localBase,link);
        }
        if(WINDOWS)
        {
            localLink=setWindows(localLink);
        }
        if(file_exists(localLink.data))
        {
            localLink=Path::merge(localBase,localLink);
            pageLinks.push(dyn(localLink));
            addLink(localLink);
        }
        else
        {
            if(DEBUG||VERBOSE)
            {
                std::cout << localLink<< " " << "not found in"<< " " << getcwd()<< "\n";
            }
        }
    _lab_parser10:;
    }
    _lab_parser11:;
    return pageLinks;
}

// parse a web page

int scanPage(text localBase,text fname)
{
    int counter=0;
    text fullpath="";
    double lmod=0;
    text ndir="";
    text cdir="";
    text link="";
    text linkBase="";
    bool ABSOLUTE=false;
    bool MYSITE=true;
    if(!parsable(fname))
    {
        return 0;
    }
    if(hasProtocol(fname))
    {
        array _I_1=localPath(fname);
        fullpath=_I_1.begin().toText();
        MYSITE=_I_1.inc().toBoolean();
        fname=getLocal(fname);
        array _I_2=Path::splitFile(fname);
        localBase=_I_2.begin().toText();
        fname=_I_2.inc().toText();
        if(MYSITE==false)
        {
            return counter;
        }
        ABSOLUTE=true;
        link=Path::merge(siteURL,localBase);
        link=Path::merge(link,fname);
        text localDir=Path::merge(localRoot,localBase);
        if(WINDOWS)
        {
            localDir=setWindows(localDir);
        }
        _sol_chdir(localDir.data);
        linkBase=localBase;
        localBase="";
    }
    else
    {
        if(getcwd()!=localRoot)
        {
            localBase=getRelativePath(getcwd());
        }
        fullpath=fname;
        if(WINDOWS)
        {
            fullpath=setWindows(fullpath);
        }
        link=createLink(siteURL,localBase,fname);
        linkBase=localBase;
    }
    cdir=getcwd();
    if(endWith(cdir,localBase))
    {
        localBase="";
    }
    if(DEBUG||VERBOSE)
    {
        std::cout << "Scanning for links"<< " " << fullpath<< " " << localBase<< " " << fname<< "\n";
    }
    if(!file_exists(fullpath.data))
    {
        display(text("   ")+fullpath+" not found");
        addList(fullpath);
        return counter;
    }
    lmod=filemtime(fullpath.data);
    if(addPage(link)==false)
    {
        if(VERBOSE==true)
        {
            std::cout << link<< " " << "already scanned"<< "\n";
        }
        return counter;
    }
    else
    {
        addTime(fullpath,lmod);
    }
    if(VERBOSE==true)
    {
        display(fullpath+":");
    }
    text k="";
    text v="";
    counter+=1;
    array dep=getLinks(linkBase,fname);
    if(dep==array())
    {
        return counter;
    }
    if(VERBOSE==true)
    {
        std::cout << dep.size()<< " " << "links found"<< "\n";
    }
    dep.reset();
    do
    {
        text page=dep.value().toText();
        text filename="";
        array _I_4=localPath(page);
        filename=_I_4.begin().toText();
        MYSITE=_I_4.inc().toBoolean();
        if(MYSITE==false)
        {
            continue;
        }
        if(filename=="")
        {
            continue;
        }
        if(inList(page)==true)
        {
            continue;
        }
        if(hasProtocol(page))
        {
            cdir=getcwd();
            _sol_chdir(basePath.data);
            counter+=scanPage("",page);
            _sol_chdir(cdir.data);
            continue;
        }
        if(Path::hasDir(filename))
        {
            array _I_5=Path::splitFile(filename);
            ndir=_I_5.begin().toText();
            filename=_I_5.inc().toText();
            if(ABSOLUTE)
            {
                ndir=Path::merge(basePath,ndir);
            }
            cdir=getcwd();
            _sol_chdir(ndir.data);
            counter+=scanPage(Path::merge(localBase,ndir),filename);
            changeDir(cdir);
        }
        else
        {
            cdir=getcwd();
            if(ABSOLUTE)
            {
                ndir=basePath;
                page=Path::merge(basePath,page);
                _sol_chdir(ndir.data);
            }
            counter+=scanPage(localBase,page);
            if(ABSOLUTE)
            {
                changeDir(cdir);
            }
        }
    }
    while(dep.bound());
    return counter;
}


void buildTag(text tagname,text value)
{
    if(value=="")
    {
        return;
    }
    output(text(text(text("     <")+tagname+">")+value+"</")+tagname+">");
    return;
}

// from the list of files
// build now a xml file

void buildTheXmlFile()
{
    text name="";
    output("<?xml version=\"1.0\" encoding=\"UTF-8\"?>");
    output("<urlset xmlns=\"http://www.google.com/schemas/sitemap/0.84\" ");
    output("xmlns:xsi=\"http://www.w3.org/2001/XMLSchema-instance\" ");
    output("xsi:schemaLocation=\"http://www.google.com/schemas/sitemap/0.84 ");
    output("http://www.google.com/schemas/sitemap/0.84/sitemap.xsd\">");
    siteList.reset();
    do
    {
        name=siteList.value().toText();
        output("  <url>");
        buildTag("loc",textToUTF8(name));
        if(PRIORITY==true)
        {
            buildTag("priority",text(0.2));
        }
        if((LASTMOD==true)||(LASTMODTIME==true))
        {
            text lmod=(siteTime.at(name)).toText();
            if(LASTMODTIME==true)
            {
                lmod=lmod.slice(0,9,1);
            }
            buildTag("lastmod",lmod);
        }
        if(FREQUENCY==true)
        {
            buildTag("changefreq",DEFAULT_FREQUENCY);
        }
        output("  </url>");
    }
    while(siteList.bound());
    output("</urlset>");
    return;
}

int levelOffset=0;
// create a tag, content UTF-8 compatible

void buildLink(text link)
{
    text linkpath="";
    text name="";
    output("<li>");
    link=textToUTF8(link);
    array _I_1=Path::splitFile(link);
    linkpath=_I_1.begin().toText();
    name=_I_1.inc().toText();
    output(text(text("<a href=")+link+">")+name+"</a>");
    output("</li>");
    return;
}


void openSubdir(text dirname)
{
    output("<br>");
    output("<h4>");
    output(dirname);
    output("</h4>");
    output("<ul>");
    return;
}


void closeSubdir()
{
    output("</ul>");
    return;
}


text removeBase(text linkpath,text base)
{
    int l=base.length();
    if(l<linkpath.length())
    {
        if(linkpath.slice(0,l,0)==base)
        {
            return linkpath.slice(l+1,-1,1);
        }
    }
    return linkpath;
}

// test if the local part of the URL holds a directory

bool hasDir(text name)
{
    int l=name.length();
    if(l<2)
    {
        return false;
    }
    if(name.find("/",1)!=-1)
    {
        return true;
    }
    return false;
}

// get dir

text getMainDir(text name)
{
    if(name.length()==0)
    {
        return "";
    }
    int i=name.find("/");
    if(i<1)
    {
        return "";
    }
    return name.slice(0,i,0);
}


void processDirs(text base)
{
    text page="";
    text name="";
    text linkpath="";
    bool empty=true;
    text currdir="";
    text thisdir="";
    for(int i=0;i<siteList.size();i++)
    {
        if((siteList.at(i))=="")
        {
            continue;
        }
        page=dyn::toText((siteList.at(i)));
        name=removeBase(page,base);
        if(!Path::hasDir(name))
        {
            buildLink(page);
            siteList[i]="";
        }
        else
        {
            empty=false;
        }
    }
    if(empty==true)
    {
        return;
    }
    for(int i=0;i<siteList.size();i++)
    {
        if((siteList.at(i))=="")
        {
            continue;
        }
        page=dyn::toText((siteList.at(i)));
        name=removeBase(page,base);
        thisdir=getMainDir(name);
        if(thisdir!=currdir)
        {
            openSubdir(thisdir);
            processDirs(text(base+"/")+thisdir);
            closeSubdir();
            currdir=thisdir;
        }
    }
    return;
}

// scan list of links
// build a text array of page in sub-dirs
// make an entry for file in current directory

void processRoot()
{
    text urlpart="";
    text name="";
    text link="";
    output("<br>");
    output("<h1>");
    output(siteURL);
    output("</h1>");
    siteList.reset();
    do
    {
        link=dyn::toText((siteList.value()));
        array _I_1=splitURL(link);
        urlpart=_I_1.begin().toText();
        name=_I_1.inc().toText();
        if(Path::hasDir(name)!=true)
        {
            buildLink(link);
            siteList[siteList.key()]="";
        }
    }
    while(siteList.bound());
    processDirs(siteURL);
    return;
}

// from the list of files
// build now a html page
// process file at the level
// then get first sub-dir, and loop

void buildTheHtmlTree()
{
    text name="";
    output("<html>");
    output("<head>");
    output("</head>");
    output("<body>");
    levelOffset=siteURL.length();
    processRoot();
    output("</body>");
    output("</html>");
    return;
}

// Main Simple Map function

int SimpleMap(int num,array args)
{
    bool changed=false;
    options(num,args);
    std::cout << version<< "\n";
    int counter=0;
    if(hasProtocol(topname))
    {
        addLink(topname);
    }
    else
    {
        addLink(Path::merge(siteURL,topname));
    }
    if(VERBOSE)
    {
        std::cout << "Web adress:"<< " " << siteURL<< "\n";
    }
    text currentPath=getcwd();
    if(currentPath!=basePath)
    {
        if(VERBOSE==true)
        {
            std::cout << "Leaving"<< " " << currentPath<< "\n";
            std::cout << "Going to"<< " " << basePath<< "\n";
        }
        changeDir(basePath);
        changed=true;
    }
    localRoot=getcwd();
    int _I_2=localRoot.length();
    for(int __2 = 0; __2 < _I_2; __2++)
    {
        text t=localRoot[__2];
        if(t=="\\")
        {
            WINDOWS=true;
            break;
        }
    }
    basePath="";
    counter+=scanPage("",topname);
    display("");
    siteList.sort();
    if(VERBOSE==true)
    {
        std::cout << "Creating"<< " " << mapname<< "\n";
    }
    
    if(mapType==0)
    {
        buildTheXmlFile();
    }
    else
    {
        if(mapType==1)
        {
            dyn::error.set(siteList.store(mapname.data));
        }
    else
    {
        if(mapType==2)
        {
            buildTheHtmlTree();
        }
    }}
    if(mapType!=1)
    {
        FILE *x=fopen(mapname.data,"w");
        siteMap.reset();
        do
        {
            text t=siteMap.value().toText();
            fwrite(x,t+"\n");
        }
        while(siteMap.bound());
        fclose(x);
    }
    dyn::error.set(logfile.store(logname.data));
    display(text(text(siteList.size())+" pages added to ")+mapname);
    if(changed==true)
    {
        if(VERBOSE==true)
        {
            std::cout << "now returning to original path"<< " " << currentPath<< "\n";
        }
        changeDir(currentPath);
    }
    return 0;
}

